"""
TODO:
 - bane
  - owned lock
  - async once

See:
 - https://github.com/davidbrochart/sqlite-anyio/blob/a3ba4c6ef0535b14a5a60071fcd6ed565a514963/sqlite_anyio/sqlite.py
 - https://github.com/rafalkrupinski/ratelimit-anyio/blob/2910a8a3d6fa54ed17ee6ba457686c9f7a4c4beb/src/ratelimit_anyio/__init__.py
 - https://github.com/nekitdev/async-extensions/tree/main/async_extensions
 - https://github.com/kinnay/anynet/tree/master/anynet
 - https://github.com/M-o-a-T/asyncscope
 - https://github.com/M-o-a-T/aevent
 - https://github.com/florimondmanca/aiometer
 - https://github.com/sanitizers/octomachinery/blob/b36c3d3d49da813ac46e361424132955a4e99ac8/octomachinery/utils/asynctools.py

==

async def killer(shutdown: anyio.Event, sleep_s: float) -> None:
    log.warning('Killing in %d seconds', sleep_s)
    await anyio.sleep(sleep_s)
    log.warning('Killing')
    shutdown.set()

"""  # noqa
import dataclasses as dc
import signal
import typing as ta

import anyio.abc
import anyio.streams.memory
import anyio.streams.stapled
import sniffio

from .. import check
from .. import lang


T = ta.TypeVar('T')

MemoryObjectReceiveStream: ta.TypeAlias = anyio.streams.memory.MemoryObjectReceiveStream
MemoryObjectSendStream: ta.TypeAlias = anyio.streams.memory.MemoryObjectSendStream

StapledByteStream: ta.TypeAlias = anyio.streams.stapled.StapledByteStream
StapledObjectStream: ta.TypeAlias = anyio.streams.stapled.StapledObjectStream


@dc.dataclass(eq=False)
class MemoryStapledObjectStream(StapledObjectStream[T]):
    send_stream: MemoryObjectSendStream[T]
    receive_stream: MemoryObjectReceiveStream[T]


##


async def anyio_eof_to_empty(fn: ta.Callable[..., ta.Awaitable[T]], *args: ta.Any, **kwargs: ta.Any) -> T | bytes:
    try:
        return await fn(*args, **kwargs)
    except anyio.EndOfStream:
        return b''


async def gather(*fns: ta.Callable[..., ta.Awaitable[T]], take_first: bool = False) -> list[lang.Maybe[T]]:
    results: list[lang.Maybe[T]] = [lang.empty()] * len(fns)

    async def inner(fn, i):
        results[i] = lang.just(await fn())
        if take_first:
            tg.cancel_scope.cancel()

    async with anyio.create_task_group() as tg:
        for i, fn in enumerate(fns):
            tg.start_soon(inner, fn, i)

    return results


async def first(*fns: ta.Callable[..., ta.Awaitable[T]], **kwargs: ta.Any) -> list[lang.Maybe[T]]:
    return await gather(*fns, take_first=True, **kwargs)


##


def get_current_task() -> anyio.TaskInfo | None:
    try:
        return anyio.get_current_task()
    except sniffio.AsyncLibraryNotFoundError:
        return None


#


BackendTask: ta.TypeAlias = ta.Union[  # noqa
    # asyncio.tasks.Task,
    # trio.lowlevel.Task,
    ta.Any,
]


def _is_class_named(obj: ta.Any, m: str, n: str) -> bool:
    cls = obj.__class__
    return cls.__module__ == m and cls.__name__ == n


def get_backend_task(at: anyio.TaskInfo) -> BackendTask | None:
    if _is_class_named(at, 'anyio._backends._asyncio', 'AsyncIOTaskInfo'):
        # https://github.com/agronholm/anyio/blob/8907964926a24461840eee0925d3f355e729f15d/src/anyio/_backends/_asyncio.py#L1846  # noqa
        # weakref.ref
        obj = at._task()  # type: ignore  # noqa
        if obj is not None and not (
                _is_class_named(obj, '_asyncio', 'Task') or
                _is_class_named(obj, 'asyncio.tasks', 'Task')
        ):
            raise TypeError(obj)
        return obj

    elif _is_class_named(at, 'anyio._backends._trio', 'TrioTaskInfo'):
        # https://github.com/agronholm/anyio/blob/8907964926a24461840eee0925d3f355e729f15d/src/anyio/_backends/_trio.py#L850  # noqa
        # weakref.proxy
        # https://stackoverflow.com/a/62144308 :|
        obj = at._task.__repr__.__self__  # type: ignore  # noqa
        if obj is not None and not _is_class_named(obj, 'trio.lowlevel', 'Task'):
            raise TypeError(obj)
        return obj

    else:
        raise TypeError(at)


def get_current_backend_task() -> BackendTask | None:
    if (at := get_current_task()) is not None:
        return get_backend_task(at)
    else:
        return None


##


def split_memory_object_streams(
        *args: anyio.create_memory_object_stream[T],
) -> tuple[
    MemoryObjectSendStream[T],
    MemoryObjectReceiveStream[T],
]:
    [tup] = args
    return tup


def create_stapled_memory_object_stream(max_buffer_size: float = 0) -> MemoryStapledObjectStream:
    return MemoryStapledObjectStream(*anyio.create_memory_object_stream(max_buffer_size))


# FIXME: https://github.com/python/mypy/issues/15238
# FIXME: https://youtrack.jetbrains.com/issues?q=tag:%20%7BPEP%20695%7D
def create_memory_object_stream[T](max_buffer_size: float = 0) -> tuple[
    MemoryObjectSendStream[T],
    MemoryObjectReceiveStream[T],
]:
    return anyio.create_memory_object_stream[T](max_buffer_size)  # noqa


def staple_memory_object_stream(
        *args: anyio.create_memory_object_stream[T],
) -> MemoryStapledObjectStream[T]:
    send, receive = args
    return MemoryStapledObjectStream(
        check.isinstance(send, MemoryObjectSendStream),  # type: ignore
        check.isinstance(receive, MemoryObjectReceiveStream),  # type: ignore
    )


# FIXME: https://github.com/python/mypy/issues/15238
# FIXME: https://youtrack.jetbrains.com/issues?q=tag:%20%7BPEP%20695%7D
def staple_memory_object_stream2[T](max_buffer_size: float = 0) -> MemoryStapledObjectStream[T]:
    send, receive = anyio.create_memory_object_stream[T](max_buffer_size)
    return MemoryStapledObjectStream(
        check.isinstance(send, MemoryObjectSendStream),  # type: ignore
        check.isinstance(receive, MemoryObjectReceiveStream),  # type: ignore
    )


##


class Once:
    def __init__(self) -> None:
        super().__init__()
        self._done = False
        self._lock = anyio.Lock()

    async def do(self, fn: ta.Callable[[], ta.Awaitable[None]]) -> bool:
        if self._done:
            return False
        async with self._lock:
            if self._done:
                return False  # type: ignore
            try:
                await fn()
            finally:
                self._done = True
            return True


class Lazy(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()
        self._once = Once()
        self._v: lang.Maybe[T] = lang.empty()

    def peek(self) -> lang.Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = lang.just(v)

    async def get(self, fn: ta.Callable[[], ta.Awaitable[T]]) -> T:
        async def do():
            self._v = lang.just(await fn())
        await self._once.do(do)
        return self._v.must()


class LazyFn(ta.Generic[T]):
    def __init__(self, fn: ta.Callable[[], ta.Awaitable[T]]) -> None:
        super().__init__()
        self._fn = fn
        self._once = Once()
        self._v: lang.Maybe[T] = lang.empty()

    def peek(self) -> lang.Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = lang.just(v)

    async def get(self) -> T:
        async def do():
            self._v = lang.just(await self._fn())
        await self._once.do(do)
        return self._v.must()


##


async def install_shutdown_signal_handler(
        tg: anyio.abc.TaskGroup,
        event: anyio.Event | None = None,
        *,
        signals: ta.Iterable[int] = (signal.SIGINT, signal.SIGTERM),
        echo: bool = False,
) -> ta.Callable[..., ta.Awaitable[None]] | None:
    if event is None:
        event = anyio.Event()

    async def _handler(*, task_status=anyio.TASK_STATUS_IGNORED):
        with anyio.open_signal_receiver(*signals) as it:  # type: ignore
            task_status.started()
            async for signum in it:
                if echo:
                    if signum == signal.SIGINT:
                        print('Ctrl+C pressed!')
                    else:
                        print('Terminated!')
                event.set()
                return

    await tg.start(_handler)
    return event.wait



import dataclasses as dc
import logging
import typing as ta

from ..lite.logs import StandardLogHandler
from ..lite.logs import configure_standard_logging as configure_lite_standard_logging
from .noisy import silence_noisy_loggers


##


FilterConfig = dict[str, ta.Any]
FormatterConfig = dict[str, ta.Any]
HandlerConfig = dict[str, ta.Any]
LoggerConfig = dict[str, ta.Any]


@dc.dataclass()
class DictConfig:
    version: int = 1
    incremental: bool = False
    disable_existing_loggers: bool = False
    filters: dict[str, FilterConfig] = dc.field(default_factory=dict)
    formatters: dict[str, FormatterConfig] = dc.field(default_factory=dict)
    handlers: dict[str, HandlerConfig] = dc.field(default_factory=dict)
    loggers: dict[str, LoggerConfig] = dc.field(default_factory=dict)
    root: LoggerConfig | None = None


##


def configure_standard_logging(
        level: ta.Any = None,
        *,
        json: bool = False,
        target: logging.Logger | None = None,
        force: bool = False,
) -> StandardLogHandler | None:
    handler = configure_lite_standard_logging(
        level,
        json=json,
        target=target,
        force=force,
    )

    if handler is None:
        return None

    #

    silence_noisy_loggers()

    #

    return handler



"""
TODO:
 - inheritance
 - dynamic registration
 - dynamic switching (skip docker if not running, skip online if not online, ...)
"""
import typing as ta

import pytest

from .... import check
from .... import collections as col
from .... import docker
from .... import lang
from ._registry import register


Configable = pytest.FixtureRequest | pytest.Config


SWITCHES: ta.Mapping[str, bool | ta.Callable[[], bool]] = {
    'docker': docker.has_cli,
    'docker-guest': docker.is_likely_in_docker,
    'online': True,
    'integration': True,
    'slow': False,
}

SWITCH_ATTRS = {k.replace('-', '_'): k for k in SWITCHES}


SwitchState: ta.TypeAlias = bool | ta.Literal['only']

SWITCH_STATE_OPT_PREFIXES: ta.Mapping[SwitchState, str] = {
    True: '--',
    False: '--no-',
    'only': '--only-',
}


def _get_obj_config(obj: Configable) -> pytest.Config:
    if isinstance(obj, pytest.Config):
        return obj
    elif isinstance(obj, pytest.FixtureRequest):
        return obj.config
    else:
        raise TypeError(obj)


def is_disabled(obj: Configable | None, name: str) -> bool:
    check.isinstance(name, str)
    check.in_(name, SWITCHES)
    return obj is not None and _get_obj_config(obj).getoption(f'--no-{name}')


def skip_if_disabled(obj: Configable | None, name: str) -> None:
    if is_disabled(obj, name):
        pytest.skip(f'{name} disabled')


def get_specified_switches(obj: Configable) -> ta.Mapping[str, SwitchState]:
    ret: dict[str, SwitchState] = {}
    for sw in SWITCHES:
        sts = {
            st
            for st, pfx in SWITCH_STATE_OPT_PREFIXES.items()
            if _get_obj_config(obj).getoption(pfx + sw)
        }
        if sts:
            if len(sts) > 1:
                raise Exception(f'Multiple switches specified for {sw}')
            ret[sw] = check.single(sts)
    return ret


@register
class SwitchesPlugin:

    def pytest_configure(self, config):
        for sw in SWITCH_ATTRS:
            config.addinivalue_line('markers', f'{sw}: mark test as {sw}')
            config.addinivalue_line('markers', f'not_{sw}: mark test as not {sw}')

    def pytest_addoption(self, parser):
        for sw in SWITCHES:
            parser.addoption(f'--no-{sw}', action='store_true', default=False, help=f'disable {sw} tests')
            parser.addoption(f'--{sw}', action='store_true', default=False, help=f'enables {sw} tests')
            parser.addoption(f'--only-{sw}', action='store_true', default=False, help=f'enables only {sw} tests')

    @lang.cached_function
    def get_switches(self) -> ta.Mapping[str, SwitchState]:
        return {
            k: v() if callable(v) else v
            for k, v in SWITCHES.items()
        }

    def pytest_collection_modifyitems(self, config, items):
        sts = {
            **self.get_switches(),
            **get_specified_switches(config),
        }

        stx = col.multi_map(map(reversed, sts.items()))  # type: ignore
        ts, fs, onlys = (stx.get(k, ()) for k in (True, False, 'only'))

        def process(item):
            sws = {sw for swa, sw in SWITCH_ATTRS.items() if swa in item.keywords}
            nsws = {sw for swa, sw in SWITCH_ATTRS.items() if ('not_' + swa) in item.keywords}

            if onlys:
                if not any(sw in onlys for sw in sws):
                    item.add_marker(pytest.mark.skip(reason=f'skipping switches {sws}'))
                    return

            else:
                for sw in sws:
                    if sw in fs:
                        item.add_marker(pytest.mark.skip(reason=f'skipping switches {sw}'))

                for nsw in nsws:
                    if nsw in ts:
                        item.add_marker(pytest.mark.skip(reason=f'skipping switches {nsw}'))

        for item in items:
            process(item)



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#


#
# This implementation of {@link ANTLRErrorListener} can be used to identify
# certain potential correctness and performance problems in grammars. "Reports"
# are made by calling {@link Parser#notifyErrorListeners} with the appropriate
# message.
#
# <ul>
# <li><b>Ambiguities</b>: These are cases where more than one path through the
# grammar can match the input.</li>
# <li><b>Weak context sensitivity</b>: These are cases where full-context
# prediction resolved an SLL conflict to a unique alternative which equaled the
# minimum alternative of the SLL conflict.</li>
# <li><b>Strong (forced) context sensitivity</b>: These are cases where the
# full-context prediction resolved an SLL conflict to a unique alternative,
# <em>and</em> the minimum alternative of the SLL conflict was found to not be
# a truly viable alternative. Two-stage parsing cannot be used for inputs where
# this situation occurs.</li>
# </ul>

from io import StringIO
from .. import Parser, DFA
from ..atn.ATNConfigSet import ATNConfigSet
from .ErrorListener import ErrorListener

class DiagnosticErrorListener(ErrorListener):

    def __init__(self, exactOnly:bool=True):
        # whether all ambiguities or only exact ambiguities are reported.
        self.exactOnly = exactOnly

    def reportAmbiguity(self, recognizer:Parser, dfa:DFA, startIndex:int,
                       stopIndex:int, exact:bool, ambigAlts:set, configs:ATNConfigSet):
        if self.exactOnly and not exact:
            return

        with StringIO() as buf:
            buf.write("reportAmbiguity d=")
            buf.write(self.getDecisionDescription(recognizer, dfa))
            buf.write(": ambigAlts=")
            buf.write(str(self.getConflictingAlts(ambigAlts, configs)))
            buf.write(", input='")
            buf.write(recognizer.getTokenStream().getText(startIndex, stopIndex))
            buf.write("'")
            recognizer.notifyErrorListeners(buf.getvalue())


    def reportAttemptingFullContext(self, recognizer:Parser, dfa:DFA, startIndex:int,
                       stopIndex:int, conflictingAlts:set, configs:ATNConfigSet):
        with StringIO() as buf:
            buf.write("reportAttemptingFullContext d=")
            buf.write(self.getDecisionDescription(recognizer, dfa))
            buf.write(", input='")
            buf.write(recognizer.getTokenStream().getText(startIndex, stopIndex))
            buf.write("'")
            recognizer.notifyErrorListeners(buf.getvalue())

    def reportContextSensitivity(self, recognizer:Parser, dfa:DFA, startIndex:int,
                       stopIndex:int, prediction:int, configs:ATNConfigSet):
        with StringIO() as buf:
            buf.write("reportContextSensitivity d=")
            buf.write(self.getDecisionDescription(recognizer, dfa))
            buf.write(", input='")
            buf.write(recognizer.getTokenStream().getText(startIndex, stopIndex))
            buf.write("'")
            recognizer.notifyErrorListeners(buf.getvalue())

    def getDecisionDescription(self, recognizer:Parser, dfa:DFA):
        decision = dfa.decision
        ruleIndex = dfa.atnStartState.ruleIndex

        ruleNames = recognizer.ruleNames
        if ruleIndex < 0 or ruleIndex >= len(ruleNames):
            return str(decision)

        ruleName = ruleNames[ruleIndex]
        if ruleName is None or len(ruleName)==0:
            return str(decision)

        return str(decision) + " (" + ruleName + ")"

    #
    # Computes the set of conflicting or ambiguous alternatives from a
    # configuration set, if that information was not already provided by the
    # parser.
    #
    # @param reportedAlts The set of conflicting or ambiguous alternatives, as
    # reported by the parser.
    # @param configs The conflicting or ambiguous configuration set.
    # @return Returns {@code reportedAlts} if it is not {@code null}, otherwise
    # returns the set of alternatives represented in {@code configs}.
    #
    def getConflictingAlts(self, reportedAlts:set, configs:ATNConfigSet):
        if reportedAlts is not None:
            return reportedAlts

        result = set()
        for config in configs:
            result.add(config.alt)

        return result



"""
TODO:
 - case
 - cast / ::
"""
import typing as ta

from ... import lang
from .base import Node
from .base import Value
from .idents import Ident
from .names import CanName
from .names import Name
from .names import NameBuilder


##


class Expr(Node, lang.Abstract):
    pass


class Literal(Expr, lang.Final):
    v: Value


class NameExpr(Expr, lang.Final):
    n: Name


class Param(Expr, lang.Final):
    n: str | None = None

    def __repr__(self) -> str:
        if self.n is not None:
            return f'{self.__class__.__name__}({self.n!r})'
        else:
            return f'{self.__class__.__name__}(@{hex(id(self))[2:]})'

    def __eq__(self, other):
        if not isinstance(other, Param):
            return False
        if self.n is None and other.n is None:
            return self is other
        else:
            return self.n == other.n


CanLiteral: ta.TypeAlias = Literal | Value
CanExpr: ta.TypeAlias = Expr | CanName | CanLiteral


class ExprBuilder(NameBuilder):
    def literal(self, o: CanLiteral) -> Literal:
        if isinstance(o, Literal):
            return o
        elif isinstance(o, Node):
            raise TypeError(o)
        else:
            return Literal(o)

    @ta.final
    def l(self, o: CanLiteral) -> Literal:  # noqa
        return self.literal(o)

    #

    def expr(self, o: CanExpr) -> Expr:
        if isinstance(o, Expr):
            return o
        elif isinstance(o, (Name, Ident)):
            return NameExpr(self.name(o))
        else:
            return self.literal(o)

    @ta.final
    def e(self, o: CanExpr) -> Expr:
        return self.expr(o)

    #

    def param(self, n: str | None = None) -> Param:
        return Param(n)

    @ta.final
    def p(self, n: str | None = None) -> Param:
        return self.param(n)



from .asyncs import (  # noqa
    AsyncConnection,
    AsyncConnectionLike,
    AsyncEngine,
    AsyncEngineLike,
    AsyncTransaction,
    AsyncTransactionLike,
    async_adapt,
)

from .exprs import (  # noqa
    paren,
)



from .. import check
from .. import reflect as rfl
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactory
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactory


class NewtypeMarshalerFactory(MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.NewType)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return ctx.make(check.isinstance(rty, rfl.NewType).ty)


class NewtypeUnmarshalerFactory(UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.NewType)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return ctx.make(check.isinstance(rty, rfl.NewType).ty)



# ruff: noqa: UP006 UP007
import dataclasses as dc
import typing as ta


DEFAULT_PYCHARM_VERSION = '242.23726.102'


@dc.dataclass(frozen=True)
class PycharmRemoteDebug:
    port: int
    host: ta.Optional[str] = 'localhost'
    install_version: ta.Optional[str] = DEFAULT_PYCHARM_VERSION


def pycharm_debug_connect(prd: PycharmRemoteDebug) -> None:
    if prd.install_version is not None:
        import subprocess
        import sys
        subprocess.check_call([
            sys.executable,
            '-mpip',
            'install',
            f'pydevd-pycharm~={prd.install_version}',
        ])

    pydevd_pycharm = __import__('pydevd_pycharm')  # noqa
    pydevd_pycharm.settrace(
        prd.host,
        port=prd.port,
        stdoutToServer=True,
        stderrToServer=True,
    )


def pycharm_debug_preamble(prd: PycharmRemoteDebug) -> str:
    import inspect
    import textwrap

    return textwrap.dedent(f"""
        {inspect.getsource(pycharm_debug_connect)}

        pycharm_debug_connect(PycharmRemoteDebug(
            {prd.port!r},
            host={prd.host!r},
            install_version={prd.install_version!r},
        ))
    """)



"""
TODO:
 - BufferedBytesGeneratorReader
 - docstrings
 - memoryviews
"""
import abc
import typing as ta

from ... import check
from .consts import DEFAULT_BUFFER_SIZE


T = ta.TypeVar('T')

O = ta.TypeVar('O')
I = ta.TypeVar('I')
R = ta.TypeVar('R')

AnyT = ta.TypeVar('AnyT', bound=ta.Any)


# Reader generators yield ints of amounts of data needed, or None for needing any amount of data.
ReaderGenerator: ta.TypeAlias = ta.Generator[int | None, I, R]

BytesReaderGenerator: ta.TypeAlias = ReaderGenerator[bytes, R]
StrReaderGenerator: ta.TypeAlias = ReaderGenerator[str, R]


# Exact reader generators always specify read sizes.
ExactReaderGenerator: ta.TypeAlias = ta.Generator[int, I, R]

BytesExactReaderGenerator: ta.TypeAlias = ExactReaderGenerator[bytes, R]
StrExactReaderGenerator: ta.TypeAlias = ExactReaderGenerator[str, R]


##


class _BytesJoiner:
    def _join(self, lst: list[bytes]) -> bytes:
        return b''.join(lst)


class _StrJoiner:
    def _join(self, lst: list[str]) -> str:
        return ''.join(lst)


##


class GeneratorReader(abc.ABC, ta.Generic[T]):
    @abc.abstractmethod
    def read(self, sz: int | None) -> ReaderGenerator[T, T]:
        raise NotImplementedError

    def read_exact(self, sz: int) -> ReaderGenerator[T, T]:
        d: ta.Any = yield from self.read(sz)
        if len(d) != sz:
            raise EOFError(f'GeneratorReader got {len(d)}, expected {sz}')
        return d


##


class PrependableGeneratorReader(GeneratorReader[AnyT]):
    def __init__(self) -> None:
        super().__init__()

        self._queue: list[tuple[AnyT, int]] = []

    @abc.abstractmethod
    def _join(self, lst: list[AnyT]) -> AnyT:
        raise NotImplementedError

    def read(self, sz: int | None) -> ReaderGenerator[AnyT, AnyT]:
        if not self._queue:
            d: AnyT = check.not_none((yield sz))
            return d

        if sz is None:
            return self._queue.pop(0)[0]

        lst: list[AnyT] = []
        rem = sz
        while rem > 0 and self._queue:
            c, p = self._queue[0]

            if len(c) - p > rem:
                lst.append(c[p:p + rem])
                self._queue[0] = (c, p + rem)
                return self._join(lst)

            lst.append(c[p:])
            rem -= len(c) - p
            self._queue.pop(0)

        if rem:
            d = check.not_none((yield rem))
            if d:
                lst.append(d)  # type: ignore[unreachable]

        if len(lst) == 1:
            return lst[0]
        else:
            return self._join(lst)

    def prepend(self, d: AnyT, p: int | None = None) -> None:
        if d:
            self._queue.insert(0, (d, p or 0))


class PrependableBytesGeneratorReader(
    _BytesJoiner,
    PrependableGeneratorReader[bytes],
):
    pass


class PrependableStrGeneratorReader(
    _StrJoiner,
    PrependableGeneratorReader[str],
):
    pass


prependable_bytes_generator_reader = PrependableBytesGeneratorReader
prependable_str_generator_reader = PrependableStrGeneratorReader


##


class BufferedGeneratorReader(PrependableGeneratorReader[AnyT], abc.ABC):
    def __init__(
            self,
            buffer_size: int = DEFAULT_BUFFER_SIZE,
    ) -> None:
        check.arg(buffer_size > 0)

        super().__init__()

        self._buffer_size = buffer_size

    def read(self, sz: int | None) -> ReaderGenerator[AnyT, AnyT]:
        g = super().read(sz)
        i: ta.Any = None
        while True:
            try:
                q = g.send(i)
            except StopIteration as e:
                return e.value

            check.state(not self._queue)

            if q is None:
                i = check.not_none((yield None))
                continue

            r = max(q, self._buffer_size)
            d: AnyT = check.not_none((yield r))
            if len(d) < q:
                i = d
                continue

            i = d[:q]
            self.prepend(d, q)


class BufferedBytesGeneratorReader(
    _BytesJoiner,
    BufferedGeneratorReader[bytes],
    PrependableGeneratorReader[bytes],
):
    pass


class BufferedStrGeneratorReader(
    _StrJoiner,
    BufferedGeneratorReader[str],
    PrependableGeneratorReader[str],
):
    pass


buffered_bytes_generator_reader = BufferedBytesGeneratorReader
buffered_str_generator_reader = BufferedStrGeneratorReader



"""
TODO:
 - integrate / expose with collections.cache
 - weakrefs (selectable by arg)
 - locks
"""
import dataclasses as dc
import functools
import inspect
import typing as ta

from .contextmanagers import DefaultLockable
from .contextmanagers import default_lock
from .descriptors import unwrap_func
from .descriptors import unwrap_func_with_partials


P = ta.ParamSpec('P')
T = ta.TypeVar('T')
CallableT = ta.TypeVar('CallableT', bound=ta.Callable)

_IGNORE = object()


def _nullary_cache_keyer():
    return ()


def _simple_cache_keyer(*args, **kwargs):
    return (args, tuple(sorted(kwargs.items())))


def _make_cache_keyer(fn, *, simple=False, bound=False):
    if simple:
        return _simple_cache_keyer

    fn, partials = unwrap_func_with_partials(fn)

    if inspect.isgeneratorfunction(fn) or inspect.iscoroutinefunction(fn):
        raise TypeError(fn)

    sig = inspect.signature(fn)
    sig_params = list(sig.parameters.values())[1 if bound else 0:]
    if not sig_params:
        return _nullary_cache_keyer

    ns = {}
    src_params = []
    src_vals = []
    kwargs_name = None
    render_pos_only_separator = False
    render_kw_only_separator = True
    for p in sig_params:
        formatted = p.name
        if p.default is not inspect.Parameter.empty:
            ns[p.name] = p.default
            formatted = f'{formatted}={formatted}'
        kind = p.kind
        if kind == inspect.Parameter.VAR_POSITIONAL:
            formatted = '*' + formatted
        elif kind == inspect.Parameter.VAR_KEYWORD:
            formatted = '**' + formatted
        if kind == inspect.Parameter.POSITIONAL_ONLY:
            render_pos_only_separator = True
        elif render_pos_only_separator:
            src_params.append('/')
            render_pos_only_separator = False
        if kind == inspect.Parameter.VAR_POSITIONAL:
            render_kw_only_separator = False
        elif kind == inspect.Parameter.KEYWORD_ONLY and render_kw_only_separator:
            src_params.append('*')
            render_kw_only_separator = False
        src_params.append(formatted)
        if kind == inspect.Parameter.VAR_KEYWORD:
            kwargs_name = p.name
        else:
            src_vals.append(p.name)
    if render_pos_only_separator:
        src_params.append('/')

    kwa = f', __builtins__.tuple(__builtins__.sorted({kwargs_name}.items()))' if kwargs_name else ''
    rendered = (
        f'def __func__({", ".join(src_params)}):\n'
        f'    return ({", ".join(src_vals)}{kwa})\n'
    )
    exec(rendered, ns)

    kfn = ns['__func__']
    for part in partials[::-1]:
        kfn = functools.partial(kfn, *part.args, **part.keywords)

    return kfn


class _CachedFunction(ta.Generic[T]):
    @dc.dataclass(frozen=True)
    class Opts:
        map_maker: ta.Callable[[], ta.MutableMapping] = dict
        simple_key: bool = False
        lock: DefaultLockable = None

    def __init__(
            self,
            fn: ta.Callable[P, T],
            *,
            opts: Opts = Opts(),
            keyer: ta.Callable[..., tuple] | None = None,
            values: ta.MutableMapping | None = None,
            value_fn: ta.Callable[P, T] | None = None,
    ) -> None:
        super().__init__()

        self._fn = (fn,)
        self._opts = opts
        self._keyer = keyer if keyer is not None else _make_cache_keyer(fn, simple=opts.simple_key)

        self._lock = default_lock(opts.lock, False)() if opts.lock is not None else None
        self._values = values if values is not None else opts.map_maker()
        self._value_fn = value_fn if value_fn is not None else fn
        functools.update_wrapper(self, fn)

    @property
    def _fn(self):
        return self.__fn

    @_fn.setter
    def _fn(self, x):
        self.__fn = x

    def reset(self) -> None:
        self._values = {}

    def __bool__(self) -> bool:
        raise TypeError

    def __call__(self, *args, **kwargs) -> T:
        k = self._keyer(*args, **kwargs)

        try:
            return self._values[k]
        except KeyError:
            pass

        if self._lock is not None:
            with self._lock:
                try:
                    return self._values[k]
                except KeyError:
                    pass

                value = self._value_fn(*args, **kwargs)

        else:
            value = self._value_fn(*args, **kwargs)

        self._values[k] = value
        return value


class _CachedFunctionDescriptor(_CachedFunction[T]):

    def __init__(
            self,
            fn: ta.Callable[P, T],
            scope: ta.Any,
            *,
            instance: ta.Any = None,
            owner: ta.Any = None,
            name: str | None = None,
            **kwargs,
    ) -> None:
        super().__init__(fn, **kwargs)

        self._scope = scope
        self._instance = instance
        self._owner = owner
        self._name = name if name is not None else unwrap_func(fn).__name__
        self._bound_keyer = None

    def __get__(self, instance, owner=None):
        scope = self._scope
        if owner is self._owner and (instance is self._instance or scope is classmethod):
            return self

        fn, = self._fn
        name = self._name
        bound_fn = fn.__get__(instance, owner)
        if self._bound_keyer is None:
            self._bound_keyer = _make_cache_keyer(fn, simple=self._opts.simple_key, bound=True)

        bound = self.__class__(
            fn,
            scope,
            opts=self._opts,
            instance=instance,
            owner=owner,
            name=name,
            keyer=self._bound_keyer,
            # values=None if scope is classmethod else self._values,
            value_fn=bound_fn,
        )

        if scope is classmethod and owner is not None:
            setattr(owner, name, bound)
        elif instance is not None:
            instance.__dict__[name] = bound

        return bound


def cached_function(fn=None, **kwargs):  # noqa
    if fn is None:
        return functools.partial(cached_function, **kwargs)
    opts = _CachedFunction.Opts(**kwargs)
    if isinstance(fn, staticmethod):
        return _CachedFunction(fn, opts=opts, value_fn=unwrap_func(fn))
    scope = classmethod if isinstance(fn, classmethod) else None
    return _CachedFunctionDescriptor(fn, scope, opts=opts)


def static_init(fn: CallableT) -> CallableT:
    fn = cached_function(fn)
    fn()
    return fn


##


class _CachedProperty(property):
    def __init__(
            self,
            fn,
            *,
            name=None,
            ignore_if=lambda _: False,
            clear_on_init=False,
    ):
        if isinstance(fn, property):
            fn = fn.fget
        super().__init__(fn)
        self._fn = fn
        self._ignore_if = ignore_if
        self._name = name
        self._clear_on_init = clear_on_init

    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self._name is None:
            raise TypeError(self)

        try:
            return instance.__dict__[self._name]
        except KeyError:
            pass

        value = self._fn.__get__(instance, owner)()
        if value is _IGNORE:
            return None
        instance.__dict__[self._name] = value
        return value

    def __set__(self, instance, value):
        if self._ignore_if(value):
            return
        if instance.__dict__[self._name] == value:
            return
        raise TypeError(self._name)

    def __delete__(self, instance):
        raise TypeError


def cached_property(fn=None, **kwargs):  # noqa
    if fn is None:
        return functools.partial(cached_property, **kwargs)
    return _CachedProperty(fn, **kwargs)



"""
TODO:
 - def maybe(v: lang.Maybe[T])
"""
import collections
import threading
import typing as ta


T = ta.TypeVar('T')
SizedT = ta.TypeVar('SizedT', bound=ta.Sized)

Message: ta.TypeAlias = str | ta.Callable[..., str | None] | None

_NONE_TYPE = type(None)

_isinstance = isinstance
_issubclass = issubclass
_callable = callable


##


_CONFIG_LOCK = threading.RLock()


OnRaiseFn: ta.TypeAlias = ta.Callable[[Exception], None]
_ON_RAISE: ta.Sequence[OnRaiseFn] = []


def register_on_raise(fn: OnRaiseFn) -> None:
    global _ON_RAISE
    with _CONFIG_LOCK:
        _ON_RAISE = [*_ON_RAISE, fn]


def unregister_on_raise(fn: OnRaiseFn) -> None:
    global _ON_RAISE
    with _CONFIG_LOCK:
        _ON_RAISE = [e for e in _ON_RAISE if e != fn]


#


_ARGS_RENDERER: ta.Callable[..., str | None] | None = None


def _try_enable_args_rendering() -> bool:
    global _ARGS_RENDERER
    if _ARGS_RENDERER is not None:
        return True

    try:
        from .diag.asts import ArgsRenderer

        ArgsRenderer.smoketest()

    except Exception:  # noqa
        return False

    def _real_render_args(fmt: str, *args: ta.Any) -> str | None:
        ra = ArgsRenderer(back=3).render_args(*args)
        if ra is None:
            return None

        return fmt % tuple(str(a) for a in ra)

    _ARGS_RENDERER = _real_render_args
    return True


_TRIED_ENABLED_ARGS_RENDERING: bool | None = None


def try_enable_args_rendering() -> bool:
    global _TRIED_ENABLED_ARGS_RENDERING
    if _TRIED_ENABLED_ARGS_RENDERING is not None:
        return _TRIED_ENABLED_ARGS_RENDERING

    with _CONFIG_LOCK:
        if _TRIED_ENABLED_ARGS_RENDERING is None:
            _TRIED_ENABLED_ARGS_RENDERING = _try_enable_args_rendering()

        return _TRIED_ENABLED_ARGS_RENDERING


##


def _default_exception_factory(exc_cls: type[Exception], *args, **kwargs) -> Exception:
    return exc_cls(*args, **kwargs)  # noqa


_EXCEPTION_FACTORY = _default_exception_factory


class _ArgsKwargs:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


def _raise(
        exception_type: type[Exception],
        default_message: str,
        message: Message,
        ak: _ArgsKwargs = _ArgsKwargs(),
        *,
        render_fmt: str | None = None,
) -> ta.NoReturn:
    exc_args = ()
    if _callable(message):
        message = ta.cast(ta.Callable, message)(*ak.args, **ak.kwargs)
        if _isinstance(message, tuple):
            message, *exc_args = message  # type: ignore

    if message is None:
        message = default_message

    try_enable_args_rendering()

    if render_fmt is not None and _ARGS_RENDERER is not None:
        rendered_args = _ARGS_RENDERER(render_fmt, *ak.args)
        if rendered_args is not None:
            message = f'{message} : {rendered_args}'

    exc = _EXCEPTION_FACTORY(
        exception_type,
        message,
        *exc_args,
        *ak.args,
        **ak.kwargs,
    )

    for fn in _ON_RAISE:
        fn(exc)

    raise exc


##


def _unpack_isinstance_spec(spec: ta.Any) -> tuple:
    if _isinstance(spec, type):
        return (spec,)
    if not _isinstance(spec, tuple):
        spec = (spec,)
    if None in spec:
        spec = tuple(filter(None, spec)) + (_NONE_TYPE,)  # noqa
    if ta.Any in spec:
        spec = (object,)
    return spec


def isinstance(v: ta.Any, spec: type[T] | tuple, msg: Message = None) -> T:  # noqa
    if not _isinstance(v, _unpack_isinstance_spec(spec)):
        _raise(
            TypeError,
            'Must be instance',
            msg,
            _ArgsKwargs(v, spec),
            render_fmt='not isinstance(%s, %s)',
        )

    return v


def of_isinstance(spec: type[T] | tuple, msg: Message = None) -> ta.Callable[[ta.Any], T]:
    def inner(v):
        return isinstance(v, _unpack_isinstance_spec(spec), msg)

    return inner


def cast(v: ta.Any, cls: type[T], msg: Message = None) -> T:  # noqa
    if not _isinstance(v, cls):
        _raise(
            TypeError,
            'Must be instance',
            msg,
            _ArgsKwargs(v, cls),
        )

    return v


def of_cast(cls: type[T], msg: Message = None) -> ta.Callable[[T], T]:
    def inner(v):
        return isinstance(v, cls, msg)

    return inner


def not_isinstance(v: T, spec: ta.Any, msg: Message = None) -> T:  # noqa
    if _isinstance(v, _unpack_isinstance_spec(spec)):
        _raise(
            TypeError,
            'Must not be instance',
            msg,
            _ArgsKwargs(v, spec),
            render_fmt='isinstance(%s, %s)',
        )

    return v


def of_not_isinstance(spec: ta.Any, msg: Message = None) -> ta.Callable[[T], T]:
    def inner(v):
        return not_isinstance(v, _unpack_isinstance_spec(spec), msg)

    return inner


##


def issubclass(v: type[T], spec: ta.Any, msg: Message = None) -> type[T]:  # noqa
    if not _issubclass(v, spec):
        _raise(
            TypeError,
            'Must be subclass',
            msg,
            _ArgsKwargs(v, spec),
            render_fmt='not issubclass(%s, %s)',
        )

    return v


def not_issubclass(v: type[T], spec: ta.Any, msg: Message = None) -> type[T]:  # noqa
    if _issubclass(v, spec):
        _raise(
            TypeError,
            'Must not be subclass',
            msg,
            _ArgsKwargs(v, spec),
            render_fmt='issubclass(%s, %s)',
        )

    return v


##


def in_(v: T, c: ta.Container[T], msg: Message = None) -> T:
    if v not in c:
        _raise(
            ValueError,
            'Must be in',
            msg,
            _ArgsKwargs(v, c),
            render_fmt='%s not in %s',
        )

    return v


def not_in(v: T, c: ta.Container[T], msg: Message = None) -> T:
    if v in c:
        _raise(
            ValueError,
            'Must not be in',
            msg,
            _ArgsKwargs(v, c),
            render_fmt='%s in %s',
        )

    return v


def empty(v: SizedT, msg: Message = None) -> SizedT:
    if len(v) != 0:
        _raise(
            ValueError,
            'Must be empty',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )

    return v


def iterempty(v: ta.Iterable[T], msg: Message = None) -> ta.Iterable[T]:
    it = iter(v)
    try:
        next(it)
    except StopIteration:
        pass
    else:
        _raise(
            ValueError,
            'Must be empty',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )

    return v


def not_empty(v: SizedT, msg: Message = None) -> SizedT:
    if len(v) == 0:
        _raise(
            ValueError,
            'Must not be empty',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )

    return v


def unique(it: ta.Iterable[T], msg: Message = None) -> ta.Iterable[T]:
    dupes = [e for e, c in collections.Counter(it).items() if c > 1]
    if dupes:
        _raise(
            ValueError,
            'Must be unique',
            msg,
            _ArgsKwargs(it, dupes),
        )

    return it


def single(obj: ta.Iterable[T], message: Message = None) -> T:
    try:
        [value] = obj
    except ValueError:
        _raise(
            ValueError,
            'Must be single',
            message,
            _ArgsKwargs(obj),
            render_fmt='%s',
        )

    return value


def opt_single(obj: ta.Iterable[T], message: Message = None) -> T | None:
    it = iter(obj)
    try:
        value = next(it)
    except StopIteration:
        return None

    try:
        next(it)
    except StopIteration:
        return value  # noqa

    _raise(
        ValueError,
        'Must be empty or single',
        message,
        _ArgsKwargs(obj),
        render_fmt='%s',
    )


##


def none(v: ta.Any, msg: Message = None) -> None:
    if v is not None:
        _raise(
            ValueError,
            'Must be None',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )


def not_none(v: T | None, msg: Message = None) -> T:
    if v is None:
        _raise(
            ValueError,
            'Must not be None',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )

    return v


##


def equal(v: T, o: ta.Any, msg: Message = None) -> T:
    if o != v:
        _raise(
            ValueError,
            'Must be equal',
            msg,
            _ArgsKwargs(v, o),
            render_fmt='%s != %s',
        )

    return v


def is_(v: T, o: ta.Any, msg: Message = None) -> T:
    if o is not v:
        _raise(
            ValueError,
            'Must be the same',
            msg,
            _ArgsKwargs(v, o),
            render_fmt='%s is not %s',
        )

    return v


def is_not(v: T, o: ta.Any, msg: Message = None) -> T:
    if o is v:
        _raise(
            ValueError,
            'Must not be the same',
            msg,
            _ArgsKwargs(v, o),
            render_fmt='%s is %s',
        )

    return v


def callable(v: T, msg: Message = None) -> T:  # noqa
    if not _callable(v):
        _raise(
            TypeError,
            'Must be callable',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )

    return v  # type: ignore


def non_empty_str(v: str | None, msg: Message = None) -> str:
    if not _isinstance(v, str) or not v:
        _raise(
            ValueError,
            'Must be non-empty str',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )

    return v


def replacing(expected: ta.Any, old: ta.Any, new: T, msg: Message = None) -> T:
    if old != expected:
        _raise(
            ValueError,
            'Must be replacing',
            msg,
            _ArgsKwargs(expected, old, new),
            render_fmt='%s -> %s -> %s',
        )

    return new


def replacing_none(old: ta.Any, new: T, msg: Message = None) -> T:
    if old is not None:
        _raise(
            ValueError,
            'Must be replacing None',
            msg,
            _ArgsKwargs(old, new),
            render_fmt='%s -> %s',
        )

    return new


##


def arg(v: bool, msg: Message = None) -> None:
    if not v:
        _raise(
            RuntimeError,
            'Argument condition not met',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )


def state(v: bool, msg: Message = None) -> None:
    if not v:
        _raise(
            RuntimeError,
            'State condition not met',
            msg,
            _ArgsKwargs(v),
            render_fmt='%s',
        )



# ruff: noqa: UP006 UP007
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import abc
import http.client
import http.server
import io
import typing as ta

from .versions import HttpProtocolVersion
from .versions import HttpProtocolVersions


T = ta.TypeVar('T')


HttpHeaders = http.client.HTTPMessage  # ta.TypeAlias


##


class ParseHttpRequestResult(abc.ABC):  # noqa
    __slots__ = (
        'server_version',
        'request_line',
        'request_version',
        'version',
        'headers',
        'close_connection',
    )

    def __init__(
            self,
            *,
            server_version: HttpProtocolVersion,
            request_line: str,
            request_version: HttpProtocolVersion,
            version: HttpProtocolVersion,
            headers: ta.Optional[HttpHeaders],
            close_connection: bool,
    ) -> None:
        super().__init__()

        self.server_version = server_version
        self.request_line = request_line
        self.request_version = request_version
        self.version = version
        self.headers = headers
        self.close_connection = close_connection

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(f"{a}={getattr(self, a)!r}" for a in self.__slots__)})'


class EmptyParsedHttpResult(ParseHttpRequestResult):
    pass


class ParseHttpRequestError(ParseHttpRequestResult):
    __slots__ = (
        'code',
        'message',
        *ParseHttpRequestResult.__slots__,
    )

    def __init__(
            self,
            *,
            code: http.HTTPStatus,
            message: ta.Union[str, ta.Tuple[str, str]],

            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self.code = code
        self.message = message


class ParsedHttpRequest(ParseHttpRequestResult):
    __slots__ = (
        'method',
        'path',
        'headers',
        'expects_continue',
        *[a for a in ParseHttpRequestResult.__slots__ if a != 'headers'],
    )

    def __init__(
            self,
            *,
            method: str,
            path: str,
            headers: HttpHeaders,
            expects_continue: bool,

            **kwargs: ta.Any,
    ) -> None:
        super().__init__(
            headers=headers,
            **kwargs,
        )

        self.method = method
        self.path = path
        self.expects_continue = expects_continue

    headers: HttpHeaders


#


class HttpRequestParser:
    DEFAULT_SERVER_VERSION = HttpProtocolVersions.HTTP_1_0

    # The default request version. This only affects responses up until the point where the request line is parsed, so
    # it mainly decides what the client gets back when sending a malformed request line.
    # Most web servers default to HTTP 0.9, i.e. don't send a status line.
    DEFAULT_REQUEST_VERSION = HttpProtocolVersions.HTTP_0_9

    #

    DEFAULT_MAX_LINE: int = 0x10000
    DEFAULT_MAX_HEADERS: int = 100

    #

    def __init__(
            self,
            *,
            server_version: HttpProtocolVersion = DEFAULT_SERVER_VERSION,

            max_line: int = DEFAULT_MAX_LINE,
            max_headers: int = DEFAULT_MAX_HEADERS,
    ) -> None:
        super().__init__()

        if server_version >= HttpProtocolVersions.HTTP_2_0:
            raise ValueError(f'Unsupported protocol version: {server_version}')
        self._server_version = server_version

        self._max_line = max_line
        self._max_headers = max_headers

    #

    @property
    def server_version(self) -> HttpProtocolVersion:
        return self._server_version

    #

    def _run_read_line_coro(
            self,
            gen: ta.Generator[int, bytes, T],
            read_line: ta.Callable[[int], bytes],
    ) -> T:
        sz = next(gen)
        while True:
            try:
                sz = gen.send(read_line(sz))
            except StopIteration as e:
                return e.value

    #

    def parse_request_version(self, version_str: str) -> HttpProtocolVersion:
        if not version_str.startswith('HTTP/'):
            raise ValueError(version_str)  # noqa

        base_version_number = version_str.split('/', 1)[1]
        version_number_parts = base_version_number.split('.')

        # RFC 2145 section 3.1 says there can be only one "." and
        #   - major and minor numbers MUST be treated as separate integers;
        #   - HTTP/2.4 is a lower version than HTTP/2.13, which in turn is lower than HTTP/12.3;
        #   - Leading zeros MUST be ignored by recipients.
        if len(version_number_parts) != 2:
            raise ValueError(version_number_parts)  # noqa
        if any(not component.isdigit() for component in version_number_parts):
            raise ValueError('non digit in http version')  # noqa
        if any(len(component) > 10 for component in version_number_parts):
            raise ValueError('unreasonable length http version')  # noqa

        return HttpProtocolVersion(
            int(version_number_parts[0]),
            int(version_number_parts[1]),
        )

    #

    def coro_read_raw_headers(self) -> ta.Generator[int, bytes, ta.List[bytes]]:
        raw_headers: ta.List[bytes] = []
        while True:
            line = yield self._max_line + 1
            if len(line) > self._max_line:
                raise http.client.LineTooLong('header line')
            raw_headers.append(line)
            if len(raw_headers) > self._max_headers:
                raise http.client.HTTPException(f'got more than {self._max_headers} headers')
            if line in (b'\r\n', b'\n', b''):
                break
        return raw_headers

    def read_raw_headers(self, read_line: ta.Callable[[int], bytes]) -> ta.List[bytes]:
        return self._run_read_line_coro(self.coro_read_raw_headers(), read_line)

    def parse_raw_headers(self, raw_headers: ta.Sequence[bytes]) -> HttpHeaders:
        return http.client.parse_headers(io.BytesIO(b''.join(raw_headers)))

    #

    def coro_parse(self) -> ta.Generator[int, bytes, ParseHttpRequestResult]:
        raw_request_line = yield self._max_line + 1

        # Common result kwargs

        request_line = '-'
        request_version = self.DEFAULT_REQUEST_VERSION

        # Set to min(server, request) when it gets that far, but if it fails before that the server authoritatively
        # responds with its own version.
        version = self._server_version

        headers: HttpHeaders | None = None

        close_connection = True

        def result_kwargs():
            return dict(
                server_version=self._server_version,
                request_line=request_line,
                request_version=request_version,
                version=version,
                headers=headers,
                close_connection=close_connection,
            )

        # Decode line

        if len(raw_request_line) > self._max_line:
            return ParseHttpRequestError(
                code=http.HTTPStatus.REQUEST_URI_TOO_LONG,
                message='Request line too long',
                **result_kwargs(),
            )

        if not raw_request_line:
            return EmptyParsedHttpResult(**result_kwargs())

        request_line = raw_request_line.decode('iso-8859-1').rstrip('\r\n')

        # Split words

        words = request_line.split()
        if len(words) == 0:
            return EmptyParsedHttpResult(**result_kwargs())

        # Parse and set version

        if len(words) >= 3:  # Enough to determine protocol version
            version_str = words[-1]
            try:
                request_version = self.parse_request_version(version_str)

            except (ValueError, IndexError):
                return ParseHttpRequestError(
                    code=http.HTTPStatus.BAD_REQUEST,
                    message=f'Bad request version ({version_str!r})',
                    **result_kwargs(),
                )

            if (
                    request_version < HttpProtocolVersions.HTTP_0_9 or
                    request_version >= HttpProtocolVersions.HTTP_2_0
            ):
                return ParseHttpRequestError(
                    code=http.HTTPStatus.HTTP_VERSION_NOT_SUPPORTED,
                    message=f'Invalid HTTP version ({version_str})',
                    **result_kwargs(),
                )

            version = min([self._server_version, request_version])

            if version >= HttpProtocolVersions.HTTP_1_1:
                close_connection = False

        # Verify word count

        if not 2 <= len(words) <= 3:
            return ParseHttpRequestError(
                code=http.HTTPStatus.BAD_REQUEST,
                message=f'Bad request syntax ({request_line!r})',
                **result_kwargs(),
            )

        # Parse method and path

        method, path = words[:2]
        if len(words) == 2:
            close_connection = True
            if method != 'GET':
                return ParseHttpRequestError(
                    code=http.HTTPStatus.BAD_REQUEST,
                    message=f'Bad HTTP/0.9 request type ({method!r})',
                    **result_kwargs(),
                )

        # gh-87389: The purpose of replacing '//' with '/' is to protect against open redirect attacks possibly
        # triggered if the path starts with '//' because http clients treat //path as an absolute URI without scheme
        # (similar to http://path) rather than a path.
        if path.startswith('//'):
            path = '/' + path.lstrip('/')  # Reduce to a single /

        # Parse headers

        try:
            raw_gen = self.coro_read_raw_headers()
            raw_sz = next(raw_gen)
            while True:
                buf = yield raw_sz
                try:
                    raw_sz = raw_gen.send(buf)
                except StopIteration as e:
                    raw_headers = e.value
                    break

            headers = self.parse_raw_headers(raw_headers)

        except http.client.LineTooLong as err:
            return ParseHttpRequestError(
                code=http.HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE,
                message=('Line too long', str(err)),
                **result_kwargs(),
            )

        except http.client.HTTPException as err:
            return ParseHttpRequestError(
                code=http.HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE,
                message=('Too many headers', str(err)),
                **result_kwargs(),
            )

        # Check for connection directive

        conn_type = headers.get('Connection', '')
        if conn_type.lower() == 'close':
            close_connection = True
        elif (
                conn_type.lower() == 'keep-alive' and
                version >= HttpProtocolVersions.HTTP_1_1
        ):
            close_connection = False

        # Check for expect directive

        expect = headers.get('Expect', '')
        if (
                expect.lower() == '100-continue' and
                version >= HttpProtocolVersions.HTTP_1_1
        ):
            expects_continue = True
        else:
            expects_continue = False

        # Return

        return ParsedHttpRequest(
            method=method,
            path=path,
            expects_continue=expects_continue,
            **result_kwargs(),
        )

    def parse(self, read_line: ta.Callable[[int], bytes]) -> ParseHttpRequestResult:
        return self._run_read_line_coro(self.coro_parse(), read_line)



import typing as ta

from ... import lang
from .base import Builder
from .base import Node


##


class Ident(Node, lang.Final):
    s: str


CanIdent: ta.TypeAlias = Ident | str


class IdentBuilder(Builder):
    def ident(self, o: CanIdent) -> Ident:
        if isinstance(o, Ident):
            return o
        elif isinstance(o, str):
            return Ident(o)
        else:
            raise TypeError(o)

    @ta.final
    def i(self, o: CanIdent) -> Ident:
        return self.ident(o)



"""
See:
 - https://duckdb.org/docs/api/python/overview.html
 - https://github.com/Mause/duckdb_engine/blob/0e3ea0107f81c66d50b444011d31fce22a9b902c/duckdb_engine/__init__.py
"""
import typing as ta

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as sap

from ... import lang


if ta.TYPE_CHECKING:
    import duckdb
else:
    duckdb = lang.proxy_import('duckdb')


class ConnectionWrapper:
    def __init__(self, c: 'duckdb.DuckDBPyConnection') -> None:
        super().__init__()
        self.__c = c
        self.autocommit = None
        self.closed = False

    def cursor(self) -> 'CursorWrapper':
        return CursorWrapper(self.__c, self)

    def __getattr__(self, name: str) -> ta.Any:
        return getattr(self.__c, name)

    def close(self) -> None:
        self.__c.close()
        self.closed = True


def _is_transaction_context_message(e: Exception) -> bool:
    return e.args[0] == 'TransactionContext Error: cannot rollback - no transaction is active'


class CursorWrapper:
    def __init__(
            self,
            c: 'duckdb.DuckDBPyConnection',
            connection_wrapper: 'ConnectionWrapper',
    ) -> None:
        super().__init__()
        self.__c = c
        self.__connection_wrapper = connection_wrapper

    def executemany(
        self,
        statement: str,
        parameters: ta.Sequence[ta.Mapping[str, ta.Any]] | None = None,
        context: ta.Any | None = None,
    ) -> None:
        self.__c.executemany(statement, list(parameters) if parameters else [])

    def execute(
        self,
        statement: str,
        parameters: ta.Sequence[ta.Any] | None = None,
        context: ta.Any | None = None,
    ) -> None:
        try:
            if statement.lower() == 'commit':  # this is largely for ipython-sql
                self.__c.commit()
            elif parameters is None:
                self.__c.execute(statement)
            else:
                self.__c.execute(statement, parameters)
        except RuntimeError as e:
            if e.args[0].startswith('Not implemented Error'):
                raise NotImplementedError(*e.args) from e
            elif _is_transaction_context_message(e):
                return
            else:
                raise

    @property
    def connection(self):
        return self.__connection_wrapper

    def close(self) -> None:
        pass

    def __getattr__(self, name: str) -> ta.Any:
        return getattr(self.__c, name)

    def fetchmany(self, size: int | None = None) -> list:
        if size is None:
            return self.__c.fetchmany()
        else:
            return self.__c.fetchmany(size)


class DuckdbDialect(sap.base.PGDialect):
    name = 'postgres__duckdb'
    driver = 'duckdb_engine'

    supports_statement_cache = False

    @classmethod
    def import_dbapi(cls):
        return duckdb

    def connect(self, *cargs, **cparams):
        conn = self.loaded_dbapi.connect(*cargs, **cparams)

        return ConnectionWrapper(conn)

    def do_rollback(self, connection) -> None:
        try:
            super().do_rollback(connection)
        except duckdb.TransactionException as e:
            if _is_transaction_context_message(e):
                return
            else:
                raise

    def do_begin(self, connection) -> None:
        connection.begin()

    def get_default_isolation_level(self, connection) -> ta.Any:
        raise NotImplementedError

    def _get_server_version_info(self, connection) -> tuple[int, int]:
        connection.execute(sa.text('select version()')).fetchone()
        return (8, 0)

    def _set_backslash_escapes(self, connection):
        pass


sa.dialects.registry.register(DuckdbDialect.name, __name__, DuckdbDialect.__name__)



# ruff: noqa: UP006 UP007
import typing as ta


class HttpProtocolVersion(ta.NamedTuple):
    major: int
    minor: int

    def __str__(self) -> str:
        return f'HTTP/{self.major}.{self.minor}'


class HttpProtocolVersions:
    HTTP_0_9 = HttpProtocolVersion(0, 9)
    HTTP_1_0 = HttpProtocolVersion(1, 0)
    HTTP_1_1 = HttpProtocolVersion(1, 1)
    HTTP_2_0 = HttpProtocolVersion(2, 0)



import dataclasses as dc
import inspect
import typing as ta

from ... import lang
from .exceptions import ValidationError
from .fields import field_init
from .fields import field_type
from .fields import has_default
from .fields import raise_field_validation_error
from .internals import HAS_DEFAULT_FACTORY
from .internals import POST_INIT_NAME
from .internals import FieldType
from .metadata import Init
from .metadata import Validate
from .processing import Processor
from .reflect import ClassInfo
from .utils import Namespace
from .utils import create_fn
from .utils import set_new_attribute


MISSING = dc.MISSING


##


def raise_validation_error(
        obj: ta.Any,
        fn: ta.Callable,
) -> ta.NoReturn:
    raise ValidationError(obj, fn)


##


class InitFields(ta.NamedTuple):
    all: ta.Sequence[dc.Field]
    ordered: ta.Sequence[dc.Field]
    std: ta.Sequence[dc.Field]
    kw_only: ta.Sequence[dc.Field]


def get_init_fields(fields: ta.Iterable[dc.Field], *, reorder: bool = False) -> InitFields:
    all_init_fields = [f for f in fields if field_type(f) in (FieldType.INSTANCE, FieldType.INIT)]
    ordered_init_fields = list(all_init_fields)
    if reorder:
        ordered_init_fields.sort(key=lambda f: (has_default(f), not f.kw_only))
    std_init_fields, kw_only_init_fields = (
        tuple(f1 for f1 in ordered_init_fields if f1.init and not f1.kw_only),
        tuple(f1 for f1 in ordered_init_fields if f1.init and f1.kw_only),
    )
    return InitFields(
        all=all_init_fields,
        ordered=ordered_init_fields,
        std=std_init_fields,
        kw_only=kw_only_init_fields,
    )


def init_param(f: dc.Field) -> str:
    if not has_default(f):
        default = ''
    elif f.default is not MISSING:
        default = f' = __dataclass_dflt_{f.name}__'
    elif f.default_factory is not MISSING:
        default = ' = __dataclass_HAS_DEFAULT_FACTORY__'
    return f'{f.name}: __dataclass_type_{f.name}__{default}'  # noqa


##


class InitBuilder:

    def __init__(
            self,
            info: ClassInfo,
            fields: ta.Mapping[str, dc.Field],
            has_post_init: bool,
            self_name: str,
            globals: Namespace,  # noqa
    ) -> None:
        super().__init__()

        self._info = info
        self._fields = fields
        self._has_post_init = has_post_init
        self._self_name = self_name
        self._globals = globals

    @lang.cached_function
    def build(self) -> ta.Callable:
        ifs = get_init_fields(self._fields.values(), reorder=self._info.params_extras.reorder)

        seen_default = None
        for f in ifs.std:
            if f.init:
                if has_default(f):
                    seen_default = f
                elif seen_default:
                    raise TypeError(f'non-default argument {f.name!r} follows default argument {seen_default.name!r}')

        locals: dict[str, ta.Any] = {}  # noqa

        if self._info.params_extras.generic_init:
            get_fty = lambda f: self._info.generic_replaced_field_annotations[f.name]
        else:
            get_fty = lambda f: f.type
        locals.update({f'__dataclass_type_{f.name}__': get_fty(f) for f in ifs.all})

        locals.update({
            '__dataclass_HAS_DEFAULT_FACTORY__': HAS_DEFAULT_FACTORY,
            '__dataclass_builtins_object__': object,
            '__dataclass_builtins_isinstance__': isinstance,
            '__dataclass_builtins_TypeError__': TypeError,
            '__dataclass_raise_validation_error__': raise_validation_error,
            '__dataclass_raise_field_validation_error__': raise_field_validation_error,
        })

        body_lines: list[str] = []
        for f in ifs.all:
            f_lines = field_init(
                f,
                self._info.params.frozen,
                locals,
                self._self_name,
                self._info.params.slots,
            )

            if f_lines:
                body_lines.extend(f_lines)

        if self._has_post_init:
            params_str = ','.join(f.name for f in ifs.all if field_type(f) is FieldType.INIT)
            body_lines.append(f'{self._self_name}.{POST_INIT_NAME}({params_str})')

        for i, fn in enumerate(self._info.merged_metadata.get(Validate, [])):
            if isinstance(fn, staticmethod):
                fn = fn.__func__
            cn = f'__dataclass_validate_{i}__'
            locals[cn] = fn
            csig = inspect.signature(fn)
            cas = ', '.join(p.name for p in csig.parameters.values())
            body_lines.append(f'if not {cn}({cas}): __dataclass_raise_validation_error__({self._self_name}, {cn})')

        inits = self._info.merged_metadata.get(Init, [])
        mro_dct = lang.build_mro_dict(self._info.cls)
        mro_v_ids = set(map(id, mro_dct.values()))
        props_by_fget_id = {id(v.fget): v for v in mro_dct.values() if isinstance(v, property) and v.fget is not None}
        for i, obj in enumerate(inits):
            if (obj_id := id(obj)) not in mro_v_ids and obj_id in props_by_fget_id:
                obj = props_by_fget_id[obj_id].__get__
            elif isinstance(obj, property):
                obj = obj.__get__
            cn = f'__dataclass_init_{i}__'
            locals[cn] = obj
            body_lines.append(f'{cn}({self._self_name})')

        if not body_lines:
            body_lines = ['pass']

        _init_params = [init_param(f) for f in ifs.std]
        if ifs.kw_only:
            _init_params += ['*']
            _init_params += [init_param(f) for f in ifs.kw_only]

        return create_fn(
            '__init__',
            [self._self_name, *_init_params],
            body_lines,
            locals=locals,
            globals=self._globals,
            return_type=lang.just(None),
        )


class InitProcessor(Processor):
    def _process(self) -> None:
        if not self._info.params.init:
            return

        has_post_init = hasattr(self._cls, POST_INIT_NAME)
        self_name = '__dataclass_self__' if 'self' in self._info.fields else 'self'

        init = InitBuilder(
            ClassInfo(self._cls),
            self._info.fields,
            has_post_init,
            self_name,
            self._info.globals,
        ).build()
        set_new_attribute(
            self._cls,
            '__init__',
            init,
        )



# ruff: noqa: UP006 UP007
import typing as ta


T = ta.TypeVar('T')
SizedT = ta.TypeVar('SizedT', bound=ta.Sized)


def check_isinstance(v: ta.Any, spec: ta.Union[ta.Type[T], tuple]) -> T:
    if not isinstance(v, spec):
        raise TypeError(v)
    return v


def check_not_isinstance(v: T, spec: ta.Union[type, tuple]) -> T:
    if isinstance(v, spec):
        raise TypeError(v)
    return v


def check_none(v: T) -> None:
    if v is not None:
        raise ValueError(v)


def check_not_none(v: ta.Optional[T]) -> T:
    if v is None:
        raise ValueError
    return v


def check_not(v: ta.Any) -> None:
    if v:
        raise ValueError(v)
    return v


def check_non_empty_str(v: ta.Optional[str]) -> str:
    if not v:
        raise ValueError
    return v


def check_state(v: bool, msg: str = 'Illegal state') -> None:
    if not v:
        raise ValueError(msg)


def check_equal(l: T, r: T) -> T:
    if l != r:
        raise ValueError(l, r)
    return l


def check_not_equal(l: T, r: T) -> T:
    if l == r:
        raise ValueError(l, r)
    return l


def check_is(l: T, r: T) -> T:
    if l is not r:
        raise ValueError(l, r)
    return l


def check_is_not(l: T, r: ta.Any) -> T:
    if l is r:
        raise ValueError(l, r)
    return l


def check_in(v: T, c: ta.Container[T]) -> T:
    if v not in c:
        raise ValueError(v, c)
    return v


def check_not_in(v: T, c: ta.Container[T]) -> T:
    if v in c:
        raise ValueError(v, c)
    return v


def check_single(vs: ta.Iterable[T]) -> T:
    [v] = vs
    return v


def check_empty(v: SizedT) -> SizedT:
    if len(v):
        raise ValueError(v)
    return v


def check_non_empty(v: SizedT) -> SizedT:
    if not len(v):
        raise ValueError(v)
    return v



# ruff: noqa: UP007
import fcntl
import os
import signal
import typing as ta


class Pidfile:
    def __init__(self, path: str) -> None:
        super().__init__()
        self._path = path

    _f: ta.TextIO

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._path!r})'

    def __enter__(self) -> 'Pidfile':
        fd = os.open(self._path, os.O_RDWR | os.O_CREAT, 0o600)
        try:
            os.set_inheritable(fd, True)
            f = os.fdopen(fd, 'r+')
        except Exception:
            try:
                os.close(fd)
            except Exception:  # noqa
                pass
            raise
        self._f = f
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, '_f'):
            self._f.close()
            del self._f

    def try_lock(self) -> bool:
        try:
            fcntl.flock(self._f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return True
        except OSError:
            return False

    def ensure_locked(self) -> None:
        if not self.try_lock():
            raise RuntimeError('Could not get lock')

    def write(self, pid: ta.Optional[int] = None) -> None:
        self.ensure_locked()
        if pid is None:
            pid = os.getpid()
        self._f.write(f'{pid}\n')
        self._f.flush()

    def clear(self) -> None:
        self.ensure_locked()
        self._f.seek(0)
        self._f.truncate()

    def read(self) -> int:
        if self.try_lock():
            raise RuntimeError('Got lock')
        self._f.seek(0)
        return int(self._f.read())

    def kill(self, sig: int = signal.SIGTERM) -> None:
        pid = self.read()
        os.kill(pid, sig)  # FIXME: Still racy



"""
TODO:
 - metaclass:
  - cleanup confer
 - descriptors - check_type/validators don't handle setters lol
 - deep_frozen?
 - field:
  - frozen
  - pickle/transient
  - mangled
  - doc
  - derive
  - check_type
 - class
  - strict_eq
  - allow_setattr
  - mangler
 - observable
 - c/py gen
 - iterable
 - proto/jsonschema gen
 - enums
 - nodal
 - embedding? forward kwargs in general? or only for replace?

TODO refs:
 - batch up exec calls
  - https://github.com/python/cpython/commit/8945b7ff55b87d11c747af2dad0e3e4d631e62d6
 - add doc parameter to dataclasses.field
  - https://github.com/python/cpython/commit/9c7657f09914254724683d91177aed7947637be5
 - add decorator argument to make_dataclass
  - https://github.com/python/cpython/commit/3e3a4d231518f91ff2f3c5a085b3849e32f1d548
"""
